/* File name: bpf_filter.uc
 * 
 * This file contains the macros filter_Init(), filter() and the 
 * filter_count() macros which are called from the dispatch loop
 * to implement packet filtering.
 *
 * filter_Init(): This macro currently does not do anything, but has
 * been declared since it is a requirement
 *
 * filter(): This macro implements the bpf filter.
 * It implements its function by first obtaining the packet from
 * the EtherIngress microblock and then processes the packet
 * as per the filter instructions. If the packet matches the
 * filter, then an exception is raised and the packet is passed
 * to the strongARM.
 *
 * Currently, the decoding of the following bpf instruction has
 * been implemented.
        protocol srcport destport srcip destip 
 *
 * filter_count(): This macro simply keeps a count of the 
 * packets found by the filter and stores it in the variable
 * stats_t->counter variable, which can be later accessed by
 * the strongARM.
 */

#include "CountControlBlock.h"
#include "Count-internal_h.uc"
#include "bpf.h"
//////////////////////////////////////////////////////////////////////////////
// IMPORTED VARIABLES (static control block)
//////////////////////////////////////////////////////////////////////////////
#include "CountImportVars.h"

#macro filter_Init()
#endm
/*
 * filter count increment macro
 */
#macro filter_count(offs)
	.local stats_addr
	immed32[stats_addr, _COUNT_STATS_BLOCK]
	scratch[incr, --, stats_addr, offs, 1]
	.endlocal
#endm

/*
 * filter match increment macro
 */
#macro port_inc(src,dest,thisPort,off)
     .local tempOff
     alu[--,src,-,thisPort]
     br!=0[checkdest#]
     immed32[tempOff,off]
     filter_count[tempOff]
     br[end_port_inc#]
checkdest#:
     alu[--,dest,-,thisPort]
     br!=0[end_port_inc#]
     immed32[tempOff,off]
     filter_count[tempOff]
end_port_inc#:
     .endlocal
#endm

/*
 * main filter macro
 */
#macro filter()
	

  .local pc input_port retc temp _p off dst type src_ip dst_ip src_port dst_port conn_filter 
       
     /*
      * Get a pointer to shared stats struct 
      * and get type which will be used for dynamic-filtering
      */

       DL_GetInputPort[input_port]
       immed32[pc, _COUNT_STATS_BLOCK]

       xbuf_alloc[$type, 1]
       immed32 [temp, TYPE_OFF]
       scratch[read, $type, pc, temp, 1], ctx_swap
       ld_field[type, 1111, $type]
       xbuf_free[$type]

       immed32[conn_filter, 0] 

      /*
       * Get a pointer to the packet data.  We don't handle the offset 
       * correctly so the packet MUST be word aligned within the buffer.
       */

       Buf_GetData(_p, dl_buffer_handle)
       DL_GetBufferOffset(off)
       alu[_p,_p,+,off, >>3]

       xbuf_alloc($$mac_dst, 4)
       sdram[read, $$mac_dst[0], _p, 0, 2], ctx_swap
       xbuf_extract(dst, $$mac_dst, 0, 0, 4)

       /* is it broadcast? */
       .if (dst == BROADCAST)
	   immed32 [temp, BCAST]
	   filter_count[temp]
	   immed32 [temp,MCAST]
	   filter_count[temp]  /* bcast is mcast by definition */
       /* is it multicast? */
       .elif(dst & MULTICAST)
           immed32 [temp,MCAST]
           filter_count[temp]
       .else   /* it must be unicast */
           immed32 [temp,UCAST]
	   filter_count[temp]
       .endif

       /* check for IP type? */
       xbuf_extract(dst,$$mac_dst,12,0,2)
       xbuf_free[$$mac_dst]
       immed32 [temp,ETH_TYPE_IP]
       alu[--,temp,-,dst]
       br!=0[notip#]

       /* get IP header  */
       xbuf_alloc($$ip_header,8)
       sdram[read, $$ip_header[0], _p,1,4],sig_done
       ctx_arb[sdram] /* wait for the read to finish */
    
       .local proto ip_hdrlen iphdr_src iphdr_dst port srcport vc
            xbuf_extract(proto, $$ip_header, 15,0,1)
            xbuf_extract(ip_hdrlen,$$ip_header,6,0,1)
       	    alu[ip_hdrlen,ip_hdrlen,AND,0x0f] /* mask out version */
       	    immed32[temp,IPHDR]  /* if hdrlen != 20 -> ip options */ 
            alu[--,temp,-,ip_hdrlen]
            br=0[noOptions#]
            immed32 [temp,IPOPTS]
            filter_count[temp]
noOptions#:
          .if (proto == PROTO_TCP || proto == PROTO_UDP)
              
             immed32[temp,IPHDR]  /* if hdrlen != 20 -> ip options */ 
             alu[--,temp,-,ip_hdrlen]
             br!=0[IpOptions#]
             xbuf_extract(port,$$ip_header,8,20,2)  /* get dest port fromip hdr */
             xbuf_extract(srcport,$$ip_header,6,20,2) /* get src port fromip hdr */
             br [chkPorts#]
IpOptions#:
             alu_shf[ip_hdrlen,--,B,ip_hdrlen,<<2]
             sdram[read, $$ip_header0,_p,ip_hdrlen,1],sig_done
             ctx_arb[sdram] /* wait for the read to finish */
             xbuf_extract(temp,$$ip_header,0,0,1) //ports are 16bit,one reg is 32bit
  	     ld_field[port, 0011, temp]
  	     ld_field[srcport, 1100, temp]
             alu_shf[srcport,--,B,srcport,>>4]

          .endif

chkPorts#:
       /* check for icmp packet? */
          .if(proto == PROTO_ICMP)
	      immed32 [temp,ICMP]
              filter_count[temp]
       /* check for tcp packet? */
          .elif (proto == PROTO_TCP) 
	     immed32 [temp,TCP]
             filter_count[temp]
       	     .if(type == proto)
	          immed32[conn_filter, 1]  /* flag for valid conn_filter */ 
                  xbuf_extract(iphdr_src, $$ip_header, 18,0,4)
                  xbuf_extract(iphdr_dst, $$ip_header, 22,0,4)
             .endif
             /* 
              * check for TCP well-known ports
              */

             /* check for SSH */
	     port_inc[srcport,port,PORT_SSH,TCPSSH]
             /* check for FTP */
             port_inc[srcport,port,PORT_FTP,TCPFTP]	
             /* check for HTTP */
             immed32[temp,PORT_HTTP] /* can't put immed val > 5 bits into alu instr */
             port_inc[srcport,port,temp,TCPHTTP]	
             port_inc[srcport,port,PORT_TELNET,TCPTEL]           

       /* check for UDP packet? */
          .elif(proto == PROTO_UDP)
	     immed32 [temp,UDP]
             filter_count[temp]
	    .if(type == proto)
	        immed32[conn_filter, 1]  /* flag for valid conn_filter */ 
                xbuf_extract(iphdr_src, $$ip_header, 18,0,4)
                xbuf_extract(iphdr_dst, $$ip_header, 22,0,4)
            .endif
             /* 
              * check for UDP well-known ports
              */

             /* check for DNS */
             immed32[temp,PORT_DNS] /* can't put immed val > 5 bits into alu instr */
	     port_inc[srcport,port,temp,UDPDNS]
             /* check for RIP */
             immed32[temp,PORT_RIP] /* can't put immed val > 5 bits into alu instr */
	     port_inc[srcport,port,temp,UDPRIP]

       /* other....kind of ip packet */
          .else
	     immed32 [temp,OTHER]
             filter_count[temp]
           .endif
        /*
         * check for dynamic filtering 
         */
        alu[--,conn_filter,-,1]
        br!=0[no_dynamic_filter#]
        /*
         * We are here because VC trace is specified by user
         * Get the IP src and dst from struct conn
         * also src and dst ports from conn struct
         */
        xbuf_alloc[$ip_hdr, 4]
        immed32[temp,SRC_OFF]
        scratch[read, $ip_hdr0, pc, temp,4], ctx_swap
  	ld_field[src_ip, 1111, $ip_hdr[0]]
  	ld_field[dst_ip, 1111, $ip_hdr[1]]
        ld_field[src_port, 1111, $ip_hdr[2]]
        ld_field[dst_port, 1111, $ip_hdr[3]]
        xbuf_free[$ip_hdr]

        immed32[vc,1]
        .if(src_ip == 0 || dst_ip == 0 || src_port == 0 || dst_port == 0)
	    immed32[vc,0]
        .endif

        .if(src_ip != 0)
           .if(src_ip != iphdr_src) 
	     br[checkvc#]
           .endif
        .endif
	.if(dst_ip != 0)
	   .if(dst_ip != iphdr_dst)
              br[checkvc#]
           .endif
        .endif
	.if(dst_port != 0)
	   .if(dst_port != port)
	      br[checkvc#]
           .endif
        .endif
	.if(src_port != 0)
           .if(src_port != srcport)
       	      br[checkvc#]
           .endif
        .endif
        immed32[temp,MATCHED]  /* dynamic filter match count */
        filter_count[temp]

#if 0 
	immed[dl_next_block, IX_EXCEPTION]
       	DL_SetExceptionCode[retc]
	DL_SetQueueNum[input_port]  /* send to SA */
#endif

        br[dynamic_filter_done#]

checkvc#:

        .if(vc == 1)

          .if(src_ip != iphdr_dst) 
             br[notmatched#]
          .endif
          .if(dst_ip != iphdr_src)
             br[notmatched#]
          .endif
          .if(dst_port != srcport)
             br[notmatched#]
          .endif
          .if(src_port != port)
             br[notmatched#]
          .endif

          immed32[temp,MYVC]  /* VC with reversed values */
          filter_count[temp]

        .endif

notmatched#:

        immed32[temp,UNMATCHED]  /* dynamic filter NOTmatch count */
        filter_count[temp]
	 
   .endlocal

dynamic_filter_done#:
no_dynamic_filter#:
   xbuf_free[$$ip_header]
notip#: 
   .endlocal
#endm



